@@ -90,7 +90,7 @@ def pan(ctx, host, bucket, namespace, cert,
 
 
 @task
-def linux(ctx, bucket, cert_key, private_key, services, profile=None):
+def linux(ctx, bucket, cert_key, private_key, services=None, profile=None):
     """
     installs certs to /etc/ssl/certs
     installs keys to /etc/ssl/private
@@ -104,10 +104,28 @@ def linux(ctx, bucket, cert_key, private_key, services, profile=None):
     s3.download_file(bucket, cert_key, filename)
     filename = os.path.join('/etc/ssl/private', os.path.basename(private_key))
     s3.download_file(bucket, private_key, filename)
-    services = services.split(',')
-    services = [ x.strip() for x in services ]
-    for x in services:
-        ctx.run(f'systemctl reload {x}')
+    if services:
+        services = services.split(',')
+        services = [ x.strip() for x in services ]
+        for x in services:
+            ctx.run(f'systemctl reload {x}')
+
+
+@task
+def xrdp(ctx, bucket, cert_key, private_key, profile=None):
+    """
+    installs certs to /etc/ssl/certs
+    installs keys to /etc/ssl/private
+    services is a comma separated list of services to reload once
+    the keys have been updated
+    """
+    from boto3 import Session
+    session = Session(profile_name=profile)
+    s3 = session.client('s3')
+    filename = '/etc/xrdp/cert.pem'
+    s3.download_file(bucket, cert_key, filename)
+    filename = '/etc/xrdp/key.pem'
+    s3.download_file(bucket, private_key, filename)
 
 
 @task
@@ -215,4 +233,5 @@ installers_collection = Collection(
     iis,
     exchange,
     rds,
+    xrdp,
 )
