@@ -1,4 +1,5 @@
 from collections import defaultdict
+from dataclasses import dataclass
 from glob import glob
 import logging
 import os
@@ -6,7 +7,7 @@ import posixpath
 import socket
 import sys
 from tempfile import NamedTemporaryFile
-from typing import Sequence, Dict
+from typing import Sequence, Dict, List, Set, Iterable
 
 from raft import task
 import yaml
@@ -21,6 +22,21 @@ from sewer.dns_providers.route53 import Route53Dns as SewerRoute53Dns
 log = logging.getLogger(__name__)
 
 
+@dataclass
+class Route53Change:
+    name: str = None
+    values: Set[str] = None
+    domain: str = None
+    zone_id: str = None
+    action: str = 'UPSERT'
+    record_type: str = 'TXT'
+    ttl: int = 60
+
+    def __post_init__(self):
+        if self.domain and not self.domain.endswith('.'):
+            self.domain = f'{self.domain}.'
+
+
 class Route53Dns(SewerRoute53Dns):
     def __init__(self, profile=None):
         super().__init__()
@@ -28,6 +44,10 @@ class Route53Dns(SewerRoute53Dns):
         self.r53 = self.session.client('route53', config=self.aws_config)
         self.waiter = self.r53.get_waiter('resource_record_sets_changed')
         self.resource_records = defaultdict(set)
+        self.change_map: Dict[str, Route53Change] = {}
+        self.zone_ids = {}
+        self.deletes = []
+        self.undos: List[Route53Change] = []
 
     def setup(self, challenges: Sequence[Dict[str, str]]) -> Sequence[ErrataItemType]:
         for x in challenges:
@@ -35,31 +55,92 @@ class Route53Dns(SewerRoute53Dns):
             value = dns_challenge(x['key_auth'])
             challenge_domain = f'_acme-challenge.{domain_name}.'
             self.resource_records[challenge_domain].add(value)
+            change = self.change_map.setdefault(
+                challenge_domain,
+                Route53Change(
+                    name=challenge_domain,
+                    domain=domain_name,
+                    record_type='TXT',
+                    values=set(),
+                ))
+            change.values.add(f'"{value}"')
+        self.find_zone_ids()
+        self.handle_existing()
         self.create_dns_record(domain_name=None, domain_dns_value=None)
         return []
 
-    def change_batch(self, action, changes):
+    def find_zone_ids(self):
+        for x in self.change_map.values():
+            pieces = x.domain.split('.')
+            while pieces:
+                d = '.'.join(pieces)
+                log.info('[route53] finding zone id for %s', d)
+                zone_id = self.zone_ids.get(d)
+                if not zone_id:
+                    response = self.r53.list_hosted_zones_by_name(DNSName=d)
+                    for zone in response['HostedZones']:
+                        if zone['Name'] == d:
+                            zone_id = self.zone_ids[d] = zone['Id']
+                            break
+                if zone_id:
+                    x.zone_id = zone_id
+                    log.info('[route53] %s => %s', d, zone_id)
+                    break
+                pieces = pieces[1:]
+
+    def change_batch(self, changes: List[Route53Change]):
         return {
-            'Comment': f'letsencrypt dns certificate validation {action}',
+            'Comment': 'letsencrypt dns certificate validation changes',
             'Changes': [{
-                'Action': action,
+                'Action': x.action,
                 'ResourceRecordSet': {
-                    'Name': name,
-                    'Type': 'TXT',
-                    'TTL': self.ttl,
+                    'Name': x.name,
+                    'Type': x.record_type,
+                    'TTL': x.ttl,
                     'ResourceRecords': [
-                        dict(Value=f'"{value}"')
-                        for value in values
+                        dict(Value=value)
+                        for value in x.values
                     ],
                 },
-            } for name, values in changes.items()],
+            } for x in changes],
         }
 
-    def by_zone_id(self):
-        by_zone_id = defaultdict(lambda: defaultdict(set))
-        for name, values in self.resource_records.items():
-            zone_id = self._find_zone_id_for_domain(name)
-            by_zone_id[zone_id][name] = values
+    def handle_existing(self):
+        domain_records = defaultdict(list)
+        paginator = self.r53.get_paginator('list_resource_record_sets')
+        for domain, st_id in self.zone_ids.items():
+            log.info('[route53] listing zone for %s', domain)
+            rg = paginator.paginate(HostedZoneId=st_id)
+            for page in rg:
+                domain_records[st_id] += page['ResourceRecordSets']
+            log.info('[route53] found %s records', len(domain_records[st_id]))
+        for x in self.change_map.values():
+            records = domain_records[x.zone_id]
+            for record in records:
+                if record['Name'] == x.name:
+                    log.info(
+                        '[route53] found existing record for %s / %s',
+                        x.name, record['Type'])
+                    change = Route53Change(
+                        name=record['Name'],
+                        record_type=record['Type'],
+                        values={ lx["Value"] for lx in record['ResourceRecords'] },
+                        ttl=record['TTL'],
+                        action='DELETE',
+                        domain=x.domain,
+                        zone_id=x.zone_id,
+                    )
+                    if record['Type'] != x.record_type:
+                        self.deletes.append(change)
+                    else:
+                        change.action = 'UPSERT'
+                        self.undos.append(change)
+
+    @classmethod
+    def by_zone_id(cls, rg: Iterable[Route53Change]):
+        by_zone_id = defaultdict(list)
+        for x in rg:
+            by_zone_id[x.zone_id].append(x)
         return by_zone_id
 
     def wait(self, zone_id, change_id):
@@ -70,16 +151,16 @@ class Route53Dns(SewerRoute53Dns):
         ))
         log.info('[route53 / %s] change is complete', zone_id)
 
-    def create_dns_record(self, domain_name=None, domain_dns_value=None):
+    def change_and_wait(self, changes: Iterable[Route53Change]):
         result = {}
-        if domain_name and domain_dns_value:
-            self.resource_records[domain_name].add(domain_dns_value)
-        for name, values in self.resource_records.items():
-            log.info('[route53] adding TXT %s => %s', name, values)
-        for zone_id, changes in self.by_zone_id().items():
+        for x in changes:
+            log.info(
+                '[route53] %s (%s) %s => %s',
+                x.action, x.record_type, x.name, x.values)
+        for zone_id, zone_changes in self.by_zone_id(changes).items():
             response = self.r53.change_resource_record_sets(
                 HostedZoneId=zone_id,
-                ChangeBatch=self.change_batch('UPSERT', changes),
+                ChangeBatch=self.change_batch(zone_changes),
             )
             change_id = response['ChangeInfo']['Id']
             result[zone_id] = change_id
@@ -87,25 +168,32 @@ class Route53Dns(SewerRoute53Dns):
             self.wait(zone_id, change_id)
         return result
 
+    def create_dns_record(self, domain_name, domain_dns_value):
+        if domain_name and domain_dns_value:
+            change = self.change_map.get(domain_name)
+            if change:
+                change.values.add(domain_dns_value)
+        if self.deletes:
+            self.change_and_wait(self.deletes)
+        result = self.change_and_wait(self.change_map.values())
+        return result
+
     def clear(self, challenges: Sequence[Dict[str, str]]) -> Sequence[ErrataItemType]:
         self.delete_dns_record(None, None)
         return []
 
-    def delete_dns_record(self, domain_name=None, domain_dns_value=None):
-        result = {}
-        if domain_name and domain_dns_value:
-            self.resource_records[domain_name].add(domain_dns_value)
-        for name, values in self.resource_records.items():
-            log.info('[route53] removing TXT %s => %s', name, values)
-        for zone_id, changes in self.by_zone_id().items():
-            response = self.r53.change_resource_record_sets(
-                HostedZoneId=zone_id,
-                ChangeBatch=self.change_batch('DELETE', changes),
-            )
-            change_id = response['ChangeInfo']['Id']
-            result[zone_id] = change_id
-        for zone_id, change_id in result.items():
-            self.wait(zone_id, change_id)
+    def delete_dns_record(self, domain_name, domain_dns_value):
+        for x in self.change_map.values():
+            if x.action == 'UPSERT':
+                x.action = 'DELETE'
+            elif x.action == 'DELETE':
+                x.action = 'UPSERT'
+        result = self.change_and_wait(self.change_map.values())
+        for x in self.deletes:
+            x.action = 'UPSERT'
+            self.undos.append(x)
+        if self.undos:
+            self.change_and_wait(self.undos)
         return result
 
 
@@ -284,50 +372,28 @@ def save_key(bucket, ns, hostname, content, tmp_dir, profile):
     s3.put_object(Bucket=bucket, Key=s3_key, Body=content, ACL='bucket-owner-full-control')
 
 
-def pfx(ctx, certificate, key):
-    f1, f2, f3 = '/tmp/p1.crt', '/tmp/p1.key', '/tmp/p1.pfx'
-    with open(f1, 'w') as f:
-        f.write(certificate)
-    with open(f2, 'w') as f:
-        f.write(key)
-    os.chmod(f1, 0o644)
-    os.chmod(f2, 0o600)
-    ctx.run(
-        f'/usr/bin/openssl pkcs12 -export -in {f1} -inkey {f2}'
-        f' -out {f3} -passout pass:')
-    with open(f3, 'rb') as f:
-        data = f.read()
-    os.remove(f1)
-    os.remove(f2)
-    os.remove(f3)
-    return data
-
-
 def full_pfx(ctx, certificate, key, tmp_dir='/tmp', password=None):
-    cert = NamedTemporaryFile(mode='w', delete=False)  # pylint: disable=consider-using-with
-    key = NamedTemporaryFile(mode='w', delete=False)  # pylint: disable=consider-using-with
-    f_pfx = NamedTemporaryFile(mode='wb', delete=False)  # pylint: disable=consider-using-with
-    with open(cert.name, 'w') as f:
-        f.write(certificate)
-    with open(key.name, 'w') as f:
-        f.write(key)
-    os.chmod(cert.name, 0o644)
-    os.chmod(key.name, 0o600)
-    if not password:
-        ctx.run(
-            f'/usr/bin/openssl pkcs12 -export -in {cert.name} -inkey {key.name}'
-            f' -out {f_pfx.name} -passout pass:')
-    else:
-        ctx.run(
-            f'/usr/bin/openssl pkcs12 -export -in {cert.name} -inkey {key.name}'
-            f' -out {f_pfx.name} -passout env:SEDGE_PASSWORD',
-            env=dict(SEDGE_PASSWORD=password),
-        )
-    with open(f_pfx.name, 'rb') as f:
-        data = f.read()
-    os.remove(cert.name)
-    os.remove(key.name)
-    os.remove(f_pfx.name)
+    with NamedTemporaryFile(mode='w') as cert, \
+            NamedTemporaryFile(mode='w') as f_key, \
+            NamedTemporaryFile(mode='wb') as f_pfx:
+        with open(cert.name, 'w') as f:
+            f.write(certificate)
+        with open(f_key.name, 'w') as f:
+            f.write(key)
+        os.chmod(cert.name, 0o644)
+        os.chmod(f_key.name, 0o600)
+        if not password:
+            ctx.run(
+                f'/usr/bin/openssl pkcs12 -export -in {cert.name} -inkey {f_key.name}'
+                f' -out {f_pfx.name} -passout pass:')
+        else:
+            ctx.run(
+                f'/usr/bin/openssl pkcs12 -export -in {cert.name} -inkey {f_key.name}'
+                f' -out {f_pfx.name} -passout env:SEDGE_PASSWORD',
+                env=dict(SEDGE_PASSWORD=password),
+            )
+        with open(f_pfx.name, 'rb') as f:
+            data = f.read()
     return data
 
 
@@ -474,7 +540,7 @@ def save_to_s3(ctx, bucket, ns, hostname, certificate, account_key, key,
 def save_to_ssm(ctx, ns, hostname, certificate, account_key, key, profile=None):
     session = Session(profile_name=profile)
     ssm = session.client('ssm')
-    pfx_data = pfx(ctx, certificate, key)
+    pfx_data = full_pfx(ctx, certificate, key)
     hostname = hostname.replace('*', 'star')
     prefix = ns
     if not prefix.startswith('/'):
@@ -632,40 +698,20 @@ def install_cert_on_windows(
     with open(filename, 'wb') as f:
         f.write(pfx_data)
     cmdlet = """
-    function Import-PfxCertificate {
-        param([string]$filename
-          , [string]$root_store = 'localmachine'
-          , [string]$store = 'My'
-          , [security.securestring]$password = $null)
-        $pfx = new-object Security.Cryptography.X509Certificates.X509Certificate2
-        $pfx.import($filename, $password, 'Exportable,PersistKeySet,MachineKeySet')
-        $p = new-object Security.Cryptography.X509Certificates.X509Store($store, $root_store)
-        $p.open('MaxAllowed')
-        $exists = $false
-        foreach ($x in $p.certificates) {
-            if ($x.thumbprint -eq $pfx.thumbprint) {
-                Write-Host 'cert already exists in cert store'
-                $exists = $true
-            }
-        }
-        if (!($exists)) {
-            $p.add($pfx)
-        }
-        $p.close()
-        return $pfx.thumbprint
-    }
     """
-    password = ''
-    store_arg = ''
     e = {}
     if pfx_password:
-        password = "-password (ConvertTo-SecureString -f -a $env:SEDGE_PASSWORD)"
+        password = '-password (ConvertTo-SecureString -f -a $env:SEDGE_PASSWORD)'
         e['SEDGE_PASSWORD'] = pfx_password
+    else:
+        password = "-password $null"
     if store:
-        store_arg = f"-store '{store}'"
+        store_arg = rf"-certstorelocation 'cert:\localmachine\{store}'"
+    else:
+        store_arg = r'-certstorelocation cert:\localmachine\my'
     lines = [
         cmdlet,
-        f"$t = Import-PfxCertificate -filename '{filename}' {password} "
+        f"$t = Import-PfxCertificate -filepath '{filename}' {password} "
         f"{store_arg}",
         'Write-Host $t',
     ]
